# ubuntu-boot-test: cmd_uefi_shim.py: UEFI shim boot test
#
# Copyright (C) 2023 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ubuntu_boot_test.config import *
from ubuntu_boot_test.util import *
from ubuntu_boot_test.vm import VirtualMachine
import os
import subprocess
import tempfile

def register(subparsers):
  parser = subparsers.add_parser("uefi_shim",
    description="UEFI shim boot test")

  parser.add_argument("-r", "--release", required=True,
    help="Guest Ubuntu release")
  parser.add_argument("-a", "--arch", required=True, type=Arch,
    help="Guest architecture")
  parser.add_argument("packages", nargs="*",
    help="List of packages to install (instead of apt-get download)")

def execute(args):
  TEMPDIR = tempfile.TemporaryDirectory("")

  PACKAGE_SETS = {
    Arch.AMD64: set((
      # NOTE AMD64 cloud images use hybrid PC+UEFI setup
      "grub2-common",
      "grub-pc",
      "grub-pc-bin",
      "grub-efi-amd64-bin",
      "grub-efi-amd64-unsigned",
      "grub-efi-amd64-signed",
      "shim-signed"
    )),
    Arch.ARM64: set((
      "grub2-common",
      "grub-efi-arm64",
      "grub-efi-arm64-bin",
      "grub-efi-arm64-unsigned",
      "grub-efi-arm64-signed",
      "shim-signed"
    )),
  }

  EFI_SUFFIXES = {
    Arch.AMD64: "x64.efi",
    Arch.ARM64: "aa64.efi"
  }

  EFI_TARGETS = {
    Arch.AMD64: "x86_64-efi-signed",
    Arch.ARM64: "arm64-efi-signed"
  }

  # Paths of packaged loaders
  SHIM_SIGNED_PATH = f"usr/lib/shim/shim{EFI_SUFFIXES[args.arch]}.signed.latest"
  MM_SIGNED_PATH = f"usr/lib/shim/mm{EFI_SUFFIXES[args.arch]}"
  FB_SIGNED_PATH = f"usr/lib/shim/fb{EFI_SUFFIXES[args.arch]}"
  GRUB_SIGNED_PATH = f"usr/lib/grub/{EFI_TARGETS[args.arch]}/grub{EFI_SUFFIXES[args.arch]}.signed"

  # Paths of installed loaders
  SHIM_ESP_PATH = f"/boot/efi/EFI/ubuntu/shim{EFI_SUFFIXES[args.arch]}"
  GRUB_ESP_PATH = f"/boot/efi/EFI/ubuntu/grub{EFI_SUFFIXES[args.arch]}"

  # Prepare packages for install
  package_paths = prepare_packages(TEMPDIR.name, PACKAGE_SETS[args.arch], args.packages)

  # Create virtual machine
  vm = VirtualMachine(TEMPDIR.name, ubuntu_cloud_url(args.release, args.arch), args.arch, Firmware.UEFI)

  def gen_sb_key(guid, name):
    if os.path.isfile(f"{name}.key"):
        return
    pem_priv, pem_cert, esl_cert = gen_efi_signkey()
    with open(os.path.join(TEMPDIR.name, f"{name}.key"), "wb") as f:
      f.write(pem_priv)
    with open(os.path.join(TEMPDIR.name, f"{name}.pem"), "wb") as f:
      f.write(pem_cert)
    vm.write_efivar(guid, name, esl_cert, append=True)

  def sbsign_file(with_key, path):
    result = subprocess.run(["sbsign",
      "--key", os.path.join(TEMPDIR.name, f"{with_key}.key"),
      "--cert", os.path.join(TEMPDIR.name, f"{with_key}.pem"),
      "--output", path, path], capture_output=not DEBUG)
    assert result.returncode == 0, f"Failed to sign {path}"

  shim_deb_path = None
  grub_signed_deb_path = None
  for package_path in package_paths:
    if "shim-signed" in package_path:
      shim_deb_path = package_path
    if "grub-efi-" in package_path and "-signed" in package_path:
      grub_signed_deb_path = package_path

  with deb_repack_ctx(TEMPDIR.name, shim_deb_path) as ctx:
    binpath = os.path.join(ctx.dir_path, SHIM_SIGNED_PATH)
    if not is_uefica_signed(binpath):
      # Sign shim with ephemeral key
      gen_sb_key(EFI_IMAGE_SECURITY_DATABASE_GUID, "db")
      sbsign_file("db", binpath)
    # Save information needed for revocation
    shim_hash_esl = pe_to_efihash(binpath)
    shim_sbat = decode_image_sbat(TEMPDIR.name, binpath)
    # Check/sign mm and fb
    for file in (MM_SIGNED_PATH, ):
      binpath = os.path.join(ctx.dir_path, file)
      if not is_canonical_signed(binpath):
        # Sign shim with ephemeral key
        gen_sb_key(SHIM_LOCK_GUID, "MokList")
        sbsign_file("MokList", binpath)

  with deb_repack_ctx(TEMPDIR.name, grub_signed_deb_path) as ctx:
    binpath = os.path.join(ctx.dir_path, GRUB_SIGNED_PATH)
    if not is_canonical_signed(binpath):
      # Sign GRUB with ephemeral key
      gen_sb_key(SHIM_LOCK_GUID, "MokList")
      sbsign_file("MokList", binpath)
    # Save information needed for revocation
    grub_hash_esl = pe_to_efihash(binpath)
    grub_sbat = decode_image_sbat(TEMPDIR.name, binpath)

  def installnew():
    # Copy packages to VM
    vm.copy_files(package_paths, "/tmp/")
    # Install packages
    vm.run_cmd(["apt", "install", "--yes", "/tmp/*.deb"])
    # Install new GRUB
    vm.run_cmd(["grub-install", "/dev/disk/by-id/virtio-0"])
    vm.run_cmd(["update-grub"])

  def cleansigs(path):
    with maybe_gzip_ctx(path) as ctx:
      # Remove exisiting signatures from the binary
      while subprocess.run(["sbattach", "--remove", ctx.path],
                           capture_output=not DEBUG).returncode == 0:
        pass

  def strip_shim():
    # Boot as an ephemeral snapshot
    vm.start(ephemeral_snapshot=True)
    # Clean signatures
    with vm.remote_file(SHIM_ESP_PATH) as rf:
      cleansigs(rf.local_path)
    # Reboot and verify that we get "Access denied"
    vm.reboot(wait=False)
    vm.waitserial(f"shim{EFI_SUFFIXES[args.arch]}: Access Denied".encode())
    # Force shutdown VM
    vm.forceshutdown()

  def strip_grub():
    # Boot as an ephemeral snapshot
    vm.start(ephemeral_snapshot=True)
    # Clean signatures
    with vm.remote_file(GRUB_ESP_PATH) as rf:
      cleansigs(rf.local_path)
    # Reboot and wait for error from MokManager
    vm.reboot(wait=False)
    vm.waitserial(b"Verification failed: (0x1A) Security Violation")
    # Force shutdown VM
    vm.forceshutdown()

  def strip_kernel():
    # Boot as an ephemeral snapshot
    vm.start(ephemeral_snapshot=True)
    # Find all kernels
    kernel_paths = vm.run_cmd(["find", "/boot", "-name", "vmlinuz-*"]).splitlines()
    if DEBUG:
      print(f"Found kernels: {kernel_paths}")
    # Clean signatures
    for kernel_path in kernel_paths:
      with vm.remote_file(kernel_path) as rf:
        cleansigs(rf.local_path)
    # Reboot and wait for error from GRUB
    vm.reboot(wait=False)
    vm.waitserial(b"error: bad shim lock signature.") # FIXME: this will change for loader protocol
    # Force shutdown VM
    vm.forceshutdown()

  def revoke_hash_esl(varguid, varname, hash_esl, failmsg):
    bk = vm.read_efivar(varguid, varname)
    # Add binary hash to revocation list
    vm.write_efivar(varguid, varname, hash_esl, append=True)
    # Ensure it is revoked
    vm.start(ephemeral_snapshot=True, wait=False)
    vm.waitserial(failmsg)
    vm.forceshutdown()
    # Restore
    vm.write_efivar(varguid, varname, bk, append=False)

  def revoke_shim_dbx():
    revoke_hash_esl(EFI_IMAGE_SECURITY_DATABASE_GUID, "dbx", shim_hash_esl,
      f"shim{EFI_SUFFIXES[args.arch]}: Access Denied".encode())

  def revoke_grub_dbx():
    revoke_hash_esl(EFI_IMAGE_SECURITY_DATABASE_GUID, "dbx", grub_hash_esl,
      b"Verification failed: (0x1A) Security Violation")

  def revoke_grub_mlx():
    revoke_hash_esl(SHIM_LOCK_GUID, "MokListX", grub_hash_esl,
      b"Verification failed: (0x1A) Security Violation")

  def revoke_comp_sbat(cnam, cgen, failmsg):
    bk = vm.read_efivar(SHIM_LOCK_GUID, "SbatLevel")
    # Set SbatLevel for component
    sbat_level = SbatLevel(bk)
    sbat_level.set_level_for(cnam, cgen)
    vm.write_efivar(SHIM_LOCK_GUID, "SbatLevel", sbat_level.encode(), append=False)
    # Ensure it is revoked
    vm.start(ephemeral_snapshot=True, wait=False)
    vm.waitserial(failmsg)
    vm.forceshutdown()
    # Restore
    vm.write_efivar(SHIM_LOCK_GUID, "SbatLevel", bk, append=False)

  def revoke_shim_sbat():
    revoke_comp_sbat("shim", shim_sbat.get_level_for("shim") + 1,
      b"Verifiying shim SBAT data failed: Security Policy Violation")

  def revoke_grub_sbat():
    revoke_comp_sbat("grub", grub_sbat.get_level_for("grub") + 1,
      b"Verification failed: (0x1A) Security Violation")

  def mok_disable_validation():
    # Boot as an ephemeral snapshot
    vm.start(ephemeral_snapshot=True)
    # Remove GRUB signature
    with vm.remote_file(GRUB_ESP_PATH) as rf:
      cleansigs(rf.local_path)
    # Run mokutil --disable-validation
    vm.run_cmd(["sh", "-c", "\"printf '11111111\\n11111111\\n' | mokutil --disable-validation\""])
    vm.run_cmd(["sh", "-c", "\"mokutil --timeout -1\""])
    # Reboot and navigate MokManager
    vm.reboot(wait=False)
    def expecthandler(serp):
      # 1. Wait for MokManager
      # 2. Move down to "Change Secure Boot state"
      # 3. Press enter
      serp.expect(b"Perform MOK management")
      serp.send(b"\x1b[B\r\n")
      # Wait for password prompt asking for 3 random chars
      for i in range(3):
        serp.expect(b"Enter password character")
        serp.send(b"1\r\n")
      # Disable Secure Boot screen
      serp.expect(b"Disable Secure Boot")
      serp.send(b"\x1b[B\r\n")
      # Then trigger reboot
      serp.expect(b"Reboot")
      serp.send(b"\r\n")
      # MokManager can hang if you dont read the entire last screen
      serp.read(0x4000)
    vm.expectserial(expecthandler)
    # Wait for successful boot
    vm.waitboot()
    vm.shutdown()

  def mok_import():
    # Boot as an ephemeral snapshot
    vm.start(ephemeral_snapshot=True)
    # Generate key and re-sign GRUB
    pem_priv, pem_cert, esl_cert = gen_efi_signkey()
    with open(os.path.join(TEMPDIR.name, f"MokImport.key"), "wb") as f:
      f.write(pem_priv)
    with open(os.path.join(TEMPDIR.name, f"MokImport.pem"), "wb") as f:
      f.write(pem_cert)
    with open(os.path.join(TEMPDIR.name, f"MokImport.der"), "wb") as f:
      # FIXME: maybe 'gen_efi_signkey()' should return a DER directly
      f.write(esl_cert[44:])
    with vm.remote_file(GRUB_ESP_PATH) as rf:
      cleansigs(rf.local_path)
      sbsign_file("MokImport", rf.local_path)
    # Copy key to machine and run mokutil --disable-validation
    vm.copy_files([os.path.join(TEMPDIR.name, f"MokImport.der")], "/tmp/")
    vm.run_cmd(["sh", "-c", "\"printf '11111111\\n11111111\\n' | mokutil --import /tmp/MokImport.der\""])
    vm.run_cmd(["sh", "-c", "\"mokutil --timeout 111\""])
    # Reboot and navigate MokManager
    vm.reboot(wait=False)
    def expecthandler(serp):
      # Wait for timeout prompt and press space
      serp.expect(b"Booting in 111 seconds")
      serp.send(b" ")
      # Select "Enroll MOK" from menu
      serp.expect(b"Continue boot")
      serp.send(b"\x1b[B\r\n")
      # [Enroll MOK] screen
      serp.expect(b"[Enroll MOK]")
      serp.send(b"\x1b[B\r\n")
      # "Enroll the key(s)?" screen
      serp.expect(b"Enroll the key(s)?")
      serp.send(b"\x1b[B\r\n")
      # "Password:" prompt
      serp.expect(b"Password:")
      serp.send(b"11111111\r\n")
      # Trigger reboot
      serp.expect(b"Reboot")
      serp.send(b"\r\n")
      # MokManager can hang if you dont read the entire last screen
      serp.read(0x4000)
    vm.expectserial(expecthandler)
    # Wait for successful boot
    vm.waitboot()
    vm.shutdown()

  def mok_sbat_policy():
    # Boot as an ephemeral snapshot
    vm.start(ephemeral_snapshot=True)
    # Get current revocations
    automatic_revocations = vm.run_cmd(["cat", "/sys/firmware/efi/efivars/SbatLevelRT-605dab50-e046-4300-abb6-3dd810dd8b23"])
    # Ask mokutil to upgrade to "latest"
    vm.run_cmd(["mokutil", "--set-sbat-policy", "latest"])
    vm.reboot()
    # Assert that they actually updated
    latest_revocations = vm.run_cmd(["cat", "/sys/firmware/efi/efivars/SbatLevelRT-605dab50-e046-4300-abb6-3dd810dd8b23"])
    assert automatic_revocations != latest_revocations
    # Shutdown
    vm.shutdown()

  TASKS = [
    (lambda: True,
      vm.start,               "Boot and provision image"),
    (lambda: True,
      installnew,             "Install new bootloaders"),
    (lambda: True,
      vm.shutdown,            "Shut down virtual machine"),
    (lambda: True,
      vm.start,               "Boot with new bootloaders"),
    (lambda: True,
      vm.shutdown,            "Shut down virtual machine"),
    (lambda: True,
      strip_shim,             "Secure Boot: stripping shim signature"),
    (lambda: True,
      strip_grub,             "Secure Boot: stripping GRUB signature"),
    (lambda: True,
      strip_kernel,           "Secure Boot: stripping kernel signature"),
    (lambda: True,
      revoke_shim_dbx,        "Secure Boot: revoking shim via dbx"),
    (lambda: True,
      revoke_grub_dbx,        "Secure Boot: revoking GRUB via dbx"),
    (lambda: True,
      revoke_grub_mlx,        "Secure Boot: revoking GRUB via MokListX"),
    (lambda: True,
      revoke_shim_sbat,       "Secure Boot: revoking shim via SBAT"),
    (lambda: True,
      revoke_grub_sbat,       "Secure Boot: revoking GRUB via SBAT"),
    (lambda: True,
      mok_disable_validation, "MOK: Verifying --disable-validation works"),
    (lambda: True,
      mok_import,             "MOK: Verifying --import works"),
    (lambda: True,
      mok_sbat_policy,        "MOK: Verifying --set-sbat-policy works"),
  ]

  for predicate, do_task, msg in TASKS:
    if predicate():
      do_task()
      print(f"{msg} OK")



# ubuntu-boot-test: cmd_pcbios_mass.py: MAAS style netboot test (PC-BIOS)
#
# Copyright (C) 2024 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ubuntu_boot_test.config import *
from ubuntu_boot_test.net import VirtualNetwork
from ubuntu_boot_test.util import *
from ubuntu_boot_test.vm import VirtualMachine
import json
import os
import shutil
import subprocess
import tempfile

def register(subparsers):
  parser = subparsers.add_parser("pcbios_maas",
    description="MAAS style netboot test (PC-BIOS)")

  parser.add_argument("-r", "--release", required=True,
    help="Guest Ubuntu release")
  parser.add_argument("-a", "--arch", required=True, type=Arch,
    help="Guest architecture")
  parser.add_argument("packages", nargs="*",
    help="List of packages to install (instead of apt-get download)")

def execute(args):
  TEMPDIR = tempfile.TemporaryDirectory("")

  MAAS_IMAGES_URL = "http://images.maas.io/ephemeral-v3/stable"
  MAAS_STABLE_STREAM_URL = f"{MAAS_IMAGES_URL}/streams/v1/com.ubuntu.maas:stable:1:bootloader-download.json"

  PACKAGE_SET = set((
    "grub2-common",
    "grub-pc",
    "grub-pc-bin"
  ))

  # Download MAAS stable loader
  def fetch_maas_loaders():
    stream = json.loads(fetch_file(MAAS_STABLE_STREAM_URL))
    products = stream["products"]["com.ubuntu.maas.stable:1:pxelinux:pxe:amd64"]
    versions = max(products["versions"].items(), key=lambda x: x[0])[1]
    download_file(f"{MAAS_IMAGES_URL}/" + versions["items"]["syslinux"]["path"],
                  os.path.join(TEMPDIR.name, "syslinux.tar.xz"))
    os.mkdir(os.path.join(TEMPDIR.name, "syslinux"))
    runcmd(["tar", "-xf", os.path.join(TEMPDIR.name, "syslinux.tar.xz")],
      cwd=os.path.join(TEMPDIR.name, "syslinux"))

  fetch_maas_loaders()

  # Netboot directory
  netbootdir = os.path.join(TEMPDIR.name, "netboot")
  os.mkdir(netbootdir)

  # Deploy PXELINUX components to netboot server
  for file in ("pxelinux.0", "ldlinux.c32"):
    shutil.copy(
      os.path.join(TEMPDIR.name, "syslinux", file),
      os.path.join(netbootdir, file))

  # Write "pxelinux.cfg"
  NETBOOT_PXELINUXCFG = """\
DEFAULT local

LABEL local
  LOCALBOOT 0
"""
  os.mkdir(os.path.join(netbootdir, "pxelinux.cfg"))
  with open(os.path.join(netbootdir, "pxelinux.cfg", "default"), "w") as f:
    f.write(NETBOOT_PXELINUXCFG)

  # Setup netboot environment
  vnet = VirtualNetwork(TEMPDIR.name, netbootdir)
  tap = vnet.new_tap()

  # Prepare packages for install
  package_paths = prepare_packages(TEMPDIR.name, PACKAGE_SET, args.packages)

  # Create virtual machine
  vm = VirtualMachine(TEMPDIR.name, ubuntu_cloud_url(args.release, args.arch), args.arch, Firmware.PCBIOS)

  def netboot():
    vm.start(tapname=tap)
    vm.shutdown()

  def installnew():
    # Boot VM
    vm.start(tapname=tap)
    # Install new packages
    vm.copy_files(package_paths, "/tmp/")
    vm.run_cmd(["apt", "install", "--yes", "/tmp/*.deb"])
    vm.run_cmd(["grub-install", "/dev/disk/by-id/virtio-0"])
    vm.run_cmd(["update-grub"])
    # Shutdown
    vm.shutdown()

  TASKS = [
    # This is the only test we really care about here
    # Aka using the PXELINUX loader in the archive to chainload a local copy
    # of the new GRUB.
    # Testing new versions of PXELINUX is currently outside our scope.
    (installnew,                    "Deploy tested bootloaders to image"),
    (netboot,                       "PXEv4: MAAS->Tested chainload"),
  ]

  for do_task, msg in TASKS:
    do_task()
    print(f"{msg} OK")



#
# ubuntu-boot-test: net.py: Netboot testing support code
#
# Copyright (C) 2024 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ubuntu_boot_test.config import *
from ubuntu_boot_test.util import *
import atexit
import os
import socket
import subprocess

num_bridges = 0
num_taps = 0
v4prefixes = set()
v6prefixes = set()

def next_bridge_name():
  """Generate a likely unused bridge device name
  """
  global num_bridges
  no_bridge = num_bridges
  num_bridges += 1
  return f"ubt-br{no_bridge}"

def next_tap_name():
  """Generate a likely unused tap device name
  """
  global num_taps
  no_tap = num_taps
  num_taps += 1
  return f"ubt-tap{no_tap}"

class InetAddr:
  def __init__(self, family, addr, mask_bits):
    self.family = family
    self.addr = addr
    self.mask_bits = mask_bits
  def __repr__(self):
    return f"{socket.inet_ntop(self.family, self.addr)}/{self.mask_bits}"
  def addr_str(self):
    return socket.inet_ntop(self.family, self.addr)

def random_ip4_subnet():
  """Generate a random IPv4 subnet (10.R.R.0/24)
  """
  while True:
    addr = b"\x0a" + os.urandom(2) + b"\0"
    if addr not in v4prefixes:
      v4prefixes.add(addr)
      break
  return InetAddr(socket.AF_INET, addr, 24)

def random_ip6_subnet():
  """Generate a random IPv6 subnet (fdRR:RRRR:RRRR:RRRR::0/64)
  """
  while True:
    addr6 = b"\xfd" + os.urandom(7) + 8 * b"\0"
    if addr6 not in v6prefixes:
      v6prefixes.add(addr6)
      break;
  return InetAddr(socket.AF_INET6, addr6, 64)

def host_addr(subnet, host_id):
  # Make sure `host_id` fits into host bits of address
  host_bits = len(subnet.addr)*8-subnet.mask_bits
  assert int.bit_length(host_id)<=host_bits, "Host identifier too large"
  # Combine `host_id` with subnet address
  i = subnet.mask_bits // 8
  result = bytearray(subnet.addr)
  for byte in int.to_bytes(host_id, length=(host_bits+7)//8, byteorder="big"):
    result[i] |= byte
    i += 1
  return InetAddr(subnet.family, bytes(result), subnet.mask_bits)

class VirtualNetwork:
  def __init__(self, tempdir, netbootdir):
    self._tempdir = tempdir
    # Create and bring and bring upbridge
    self._br = next_bridge_name()
    runcmd(["ip", "link", "add", self._br, "type", "bridge"])
    runcmd(["ip", "link", "set", self._br, "up"])
    atexit.register(lambda: runcmd(["ip", "link", "del", self._br], assert_ok=False))
    # Create subnets
    self._v4subnet = random_ip4_subnet()
    self._v6subnet = random_ip6_subnet()
    # Create address for the host
    self._v4addr = host_addr(self._v4subnet, 1)
    self._v6addr = host_addr(self._v6subnet, 1)
    runcmd(["ip", "addr", "add", f"{self._v4addr}", "dev", self._br])
    runcmd(["ip", "addr", "add", f"{self._v6addr}", "dev", self._br])
    # Configure dnsmasq
    self._dnsmasq_conf = os.path.join(self._tempdir, "dnsmasq.conf")
    self._netbootdir = netbootdir
    with open(self._dnsmasq_conf, "w") as f:
      f.write(f"""\
# Disable DNS
port=0

# Bind DHCP and TFTP to bridge
interface={self._br}
bind-interfaces

# IPv4 client address range
dhcp-range={host_addr(self._v4subnet, 100).addr_str()},{host_addr(self._v4subnet, 200).addr_str()},12h

# IPv4 TFTP boot
dhcp-match=set:bios-tftp-x86,option:client-arch,0
dhcp-boot=tag:bios-tftp-x86,pxelinux.0
dhcp-match=set:efi-tftp-x86_64,option:client-arch,7
dhcp-boot=tag:efi-tftp-x86_64,bootx64.efi
dhcp-match=set:efi-tftp-arm64,option:client-arch,11
dhcp-boot=tag:efi-tftp-arm64,bootaa64.efi

# IPv4 HTTP boot
dhcp-match=set:efi-http-x86_64,option:client-arch,16
dhcp-option=tag:efi-http-x86_64,option:bootfile-name,http://{self._v4addr.addr_str()}:8000/bootx64.efi
dhcp-option=tag:efi-http-x86_64,option:vendor-class,HTTPClient
dhcp-match=set:efi-http-arm64,option:client-arch,19
dhcp-option=tag:efi-http-arm64,option:bootfile-name,http://{self._v4addr.addr_str()}:8000/bootaa64.efi
dhcp-option=tag:efi-http-arm64,option:vendor-class,HTTPClient

# IPv6 client address range
dhcp-range={host_addr(self._v6subnet, 0x1000).addr_str()},{host_addr(self._v6subnet, 0x2000).addr_str()},12h

# IPv6 TFTP boot
dhcp-match=set:efi-tftp-x86_64-v6,option6:61,7
dhcp-option=tag:efi-tftp-x86_64-v6,option6:bootfile-url,tftp://[{self._v6addr.addr_str()}]/bootx64.efi
dhcp-match=set:efi-tftp-arm64-v6,option6:61,11
dhcp-option=tag:efi-tftp-arm64-v6,option6:bootfile-url,tftp://[{self._v6addr.addr_str()}]/bootaa64.efi

# IPv6 HTTP boot
dhcp-match=set:efi-http-x86_64-v6,option6:61,16
dhcp-option=tag:efi-http-x86_64-v6,option6:bootfile-url,http://[{self._v6addr.addr_str()}]:8000/bootx64.efi
dhcp-option=tag:efi-http-x86_64-v6,option6:16,AA,HTTPClient
dhcp-match=set:efi-http-arm64-v6,option6:61,19
dhcp-option=tag:efi-http-arm64-v6,option6:bootfile-url,http://[{self._v6addr.addr_str()}]:8000/bootaa64.efi
dhcp-option=tag:efi-http-arm64-v6,option6:16,AA,HTTPClient

# FIXME: The line described below is a kludge to work around the seeming limitations of dnsmasq
# FIXME: The option we need to encode is 2 byte ID, 2 byte length, 4 byte enterprise ID, 2 byte length, string
# FIXME: The encoding dnsmasq does is 2 byte ID, 2 byte length, then for each value 2 byte length, string
# FIXME: Take length (2) + "AA" becomes the enterprise ID (which firmware ignores), so it all
# FIXME: works despite it being all horrible. Maybe we should use a different DHCP server.
# FIXME: dhcp-option=tag:efi-http-arm64-v6,option6:16,AA,HTTPClient

# TFTP server
enable-tftp
tftp-root={self._netbootdir}
""")
    # Start dnsmasq
    self._dnsmasq = ProcessWrapper(["dnsmasq",
      "--no-daemon", f"--conf-file={self._dnsmasq_conf}"])
    # Start HTTP servers
    import time
    time.sleep(5) # FIXME: wait for address assignment in a nicer way
    self._http_daemon_v4 = ProcessWrapper(["python3",
      "-m", "http.server",
      "-b", self._v4addr.addr_str(),
      "-d", self._netbootdir,
      "8000"])
    self._http_daemon_v6 = ProcessWrapper(["python3",
      "-m", "http.server",
      "-b", f"{self._v6addr.addr_str()}",
      "-d", self._netbootdir,
      "8000"])

  def new_tap(self):
    tap = next_tap_name()
    runcmd(["ip", "tuntap", "add", "mode", "tap", "dev", tap])
    runcmd(["ip", "link", "set", "dev", tap, "master", self._br])
    runcmd(["ip", "link", "set", "dev", tap, "up"])
    atexit.register(lambda: runcmd(["ip", "link", "del", tap], assert_ok=False))
    return tap



# ubuntu-boot-test: cmd_uefi_mass.py: MAAS style netboot test (UEFI)
#
# Copyright (C) 2024 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ubuntu_boot_test.config import *
from ubuntu_boot_test.net import VirtualNetwork
from ubuntu_boot_test.util import *
from ubuntu_boot_test.vm import VirtualMachine
import json
import os
import shutil
import subprocess
import tempfile

def register(subparsers):
  parser = subparsers.add_parser("uefi_maas",
    description="MAAS style netboot test (UEFI)")

  parser.add_argument("-r", "--release", required=True,
    help="Guest Ubuntu release")
  parser.add_argument("-a", "--arch", required=True, type=Arch,
    help="Guest architecture")
  parser.add_argument("packages", nargs="*",
    help="List of packages to install (instead of apt-get download)")

def execute(args):
  TEMPDIR = tempfile.TemporaryDirectory("")

  PACKAGE_SETS = {
    Arch.AMD64: set((
      # NOTE AMD64 cloud images use hybrid PC+UEFI setup
      "grub2-common",
      "grub-pc",
      "grub-pc-bin",
      "grub-efi-amd64-bin",
      "grub-efi-amd64-unsigned",
      "grub-efi-amd64-signed",
      "shim-signed"
    )),
    Arch.ARM64: set((
      "grub2-common",
      "grub-efi-arm64",
      "grub-efi-arm64-bin",
      "grub-efi-arm64-unsigned",
      "grub-efi-arm64-signed",
      "shim-signed"
    )),
  }

  EFI_SUFFIXES = {
    Arch.AMD64: "x64.efi",
    Arch.ARM64: "aa64.efi"
  }

  EFI_TARGETS = {
    Arch.AMD64: "x86_64-efi-signed",
    Arch.ARM64: "arm64-efi-signed"
  }

  # Paths of packaged loaders
  SHIM_SIGNED_PATH = f"usr/lib/shim/shim{EFI_SUFFIXES[args.arch]}.signed.latest"
  GRUB_SIGNED_PATH = f"usr/lib/grub/{EFI_TARGETS[args.arch]}/grub{EFI_SUFFIXES[args.arch]}.signed"
  GRUBNET_SIGNED_PATH = f"usr/lib/grub/{EFI_TARGETS[args.arch]}/grubnet{EFI_SUFFIXES[args.arch]}.signed"

  # Netboot directory
  netbootdir = os.path.join(TEMPDIR.name, "netboot")
  os.mkdir(netbootdir)
  os.mkdir(os.path.join(netbootdir, "grub"))

  # Write "grub.cfg"
  # Ubuntu loader lives at /EFI/ubuntu/shimx64.efi
  # (We cannot boot it with chainloader directly because signature verification
  #  is only done via shim in our grub when using chainloader)
  NETBOOT_GRUBCFG = f"""\
BOOTLOADER=/efi/ubuntu/shim{EFI_SUFFIXES[args.arch]}
search --set=root --file $BOOTLOADER
if [ $? -eq 0 ]; then
  chainloader $BOOTLOADER
  boot
fi
BOOTLOADER=/efi/boot/boot{EFI_SUFFIXES[args.arch]}
search --set=root --file $BOOTLOADER
if [ $? -eq 0 ]; then
  rmmod peimage
  exit
fi
"""
  with open(os.path.join(netbootdir, "grub", "grub.cfg"), "w") as f:
    f.write(NETBOOT_GRUBCFG)

  # Setup netboot environment
  vnet = VirtualNetwork(TEMPDIR.name, netbootdir)
  tap = vnet.new_tap()

  # Prepare packages for install
  package_paths = prepare_packages(TEMPDIR.name, PACKAGE_SETS[args.arch], args.packages)

  # Create virtual machine
  vm = VirtualMachine(TEMPDIR.name, ubuntu_cloud_url(args.release, args.arch), args.arch, Firmware.UEFI)

  def gen_sb_key(guid, name):
    pem_priv, pem_cert, esl_cert = gen_efi_signkey()
    with open(os.path.join(TEMPDIR.name, f"{name}.key"), "wb") as f:
      f.write(pem_priv)
    with open(os.path.join(TEMPDIR.name, f"{name}.pem"), "wb") as f:
      f.write(pem_cert)
    vm.write_efivar(guid, name, esl_cert, append=True)

  def sbsign_file(with_key, path):
    result = subprocess.run(["sbsign",
      "--key", os.path.join(TEMPDIR.name, f"{with_key}.key"),
      "--cert", os.path.join(TEMPDIR.name, f"{with_key}.pem"),
      "--output", path, path], capture_output=not DEBUG)
    assert result.returncode == 0, f"Failed to sign {path}"

  shim_deb_path = None
  grub_signed_deb_path = None
  for package_path in package_paths:
    if "shim-signed" in package_path:
      shim_deb_path = package_path
    if "grub-efi-" in package_path and "-signed" in package_path:
      grub_signed_deb_path = package_path

  with deb_repack_ctx(TEMPDIR.name, shim_deb_path) as ctx:
    binpath = os.path.join(ctx.dir_path, SHIM_SIGNED_PATH)
    if not is_uefica_signed(binpath):
      # Sign shim with ephemeral key
      gen_sb_key(EFI_IMAGE_SECURITY_DATABASE_GUID, "db")
      sbsign_file("db", binpath)
    # Copy shim to netboot directory
    shutil.copy(binpath, os.path.join(netbootdir, f"boot{EFI_SUFFIXES[args.arch]}"))

  with deb_repack_ctx(TEMPDIR.name, grub_signed_deb_path) as ctx:
    binpath = os.path.join(ctx.dir_path, GRUB_SIGNED_PATH)
    netbinpath = os.path.join(ctx.dir_path, GRUBNET_SIGNED_PATH)
    if not is_canonical_signed(binpath):
      # Sign GRUB with ephemeral key
      gen_sb_key(SHIM_LOCK_GUID, "MokList")
      sbsign_file("MokList", binpath)
      sbsign_file("MokList", netbinpath)
    # Copy GRUB to netboot directory
    shutil.copy(netbinpath, os.path.join(netbootdir, f"grub{EFI_SUFFIXES[args.arch]}"))

  def setup_netboot_type(netboot_type):
    # Iterate BootXXXX entries and disable entries we don't want
    boot_order = vm.read_efivar(EFI_GLOBAL_VARIABLE_GUID, "BootOrder")
    for i in range(0, len(boot_order), 2):
      num = int(boot_order[i+1])<<8|int(boot_order[i])
      val = bytearray(vm.read_efivar(EFI_GLOBAL_VARIABLE_GUID, f"Boot{num:04X}"))
      name = []
      for j in range(6, len(val), 2):
        if val[j] == 0:
          break
        name.append(chr(int(val[j+1]<<8|int(val[j]))))
      name = "".join(name)
      # We identify the card we want to netboot from via this hard-coded MAC,
      # it is kind of ugly, but parsing device paths is more annoying
      if netboot_type not in name or "(MAC:000000000001)" not in name:
        # Bit 0 of Attributes is LOAD_OPTION_ACTIVE
        val[0] &= 0xfe
        vm.write_efivar(EFI_GLOBAL_VARIABLE_GUID, f"Boot{num:04X}", val, append=False)

  def netboot(standard):
    setup_netboot_type(standard)
    vm.start(tapname=tap, wait=False)
    # Make sure we really are netbooting by "grep"-ing for output
    # It would be nice if we could just disable all other BootXXXX entries
    # but OVMF still tries built-in entries even when setting ACTIVE=0
    if "HTTP" in standard:
      # Shim doesn't print anything on HTTP boot, but firmware prints this
      vm.waitserial(b"URI: http://")
    else:
      # This is printed by shim on TFTP boot
      vm.waitserial(b"Fetching Netboot Image")
    vm.waitboot()
    vm.shutdown()

  def installnew():
    # Boot VM
    vm.start(tapname=tap)
    # Install new packages
    vm.copy_files(package_paths, "/tmp/")
    vm.run_cmd(["apt", "install", "--yes", "/tmp/*.deb"])
    vm.run_cmd(["grub-install", "/dev/disk/by-id/virtio-0"])
    vm.run_cmd(["update-grub"])
    # Shutdown
    vm.shutdown()

  TASKS = [
    (lambda: True,
      vm.start,                      "Boot and provision image"),
    (lambda: True,
      vm.shutdown,                   "Shut down virtual machine"),

    # Tested -> Distro
    (lambda: True,
      lambda: netboot("PXEv4"),      "PXEv4: Tested->Distro chainload"),
    (lambda: True,
      lambda: netboot("PXEv6"),      "PXEv6: Tested->Distro chainload"),
    (lambda: True,
      lambda: netboot("HTTPv4"),     "HTTPv4: Tested->Distro chainload"),
    # IPv6 HTTP boot currently only works when manually selected from the boot
    # menu, but not via "BootOrder". This seems like an edk2 bug.
    # (lambda: True,
    #   lambda: netboot("HTTPv6"),     "HTTPv6: Tested->Distro chainload"),

    # Tested -> Tested chainload
    (lambda: True,
      installnew,                    "Deploy tested bootloaders to image"),
    (lambda: True,
      lambda: netboot("PXEv4"),      "PXEv4: Tested->Tested chainload"),
    (lambda: True,
      lambda: netboot("PXEv6"),      "PXEv6: Tested->Tested chainload"),
    (lambda: True,
      lambda: netboot("HTTPv4"),     "HTTPv4: Tested->Tested chainload"),
    # (lambda: True,
    #   lambda: netboot("HTTPv6"),   "HTTPv6: Tested->Tested chainload"),
  ]

  for predicate, do_task, msg in TASKS:
    if predicate():
      do_task()
      print(f"{msg} OK")



#
# ubuntu-boot-test: util.py: Utility classes and functions
#
# Copyright (C) 2023 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from enum import Enum
from OpenSSL import crypto
from ubuntu_boot_test.config import *
import atexit
import binascii
import gzip
import os
import requests
import shutil
import signal
import struct
import subprocess
import sys


class Arch(Enum):
  """Processor architecture
  """
  AMD64   = "amd64"
  ARM64   = "arm64"
  PPC64EL = "ppc64el"

  def __str__(self):
    return self.value

class Firmware(Enum):
  """Firmware type
  """
  UEFI     = "uefi"
  PCBIOS   = "pcbios"
  IEEE1275 = "ieee1275"

  def __str__(self):
    return self.value

def host_kernel_arch():
  """Determine the host's kernel architecture
  """
  machine = os.uname().machine
  if machine == "x86_64":
    return Arch.AMD64
  if machine == "aarch64":
    return Arch.ARM64
  if machine == "ppc64le": # NOTE: Linux uses le instead of el
    return Arch.PPC64EL
  assert False, f"Unknown kernel architecture {machine}"

def kvm_supported(guest_arch):
  # Workaround: /dev/kvm is exposed on ppc64el autopkgtest testbeds, but nested kvm doesnt actually work
  if guest_arch == Arch.PPC64EL:
    return False
  # Accelaration can be used when the guest has the same architecture as the
  # host kernel (NOTE: this is a simplification but it's good enough here)
  # We also need to ensure that the KVM device node is exposed and accessible
  if host_kernel_arch() == guest_arch and \
      os.access("/dev/kvm", os.R_OK | os.W_OK):
    return True
  return False

def ubuntu_cloud_url(release, arch):
  arch = { Arch.AMD64: "amd64",
           Arch.ARM64: "arm64",
           Arch.PPC64EL: "ppc64el" }[arch]

  if LOCAL_IMG_SERVER:
    return f"http://localhost/cloudimg/{release}-server-cloudimg-{arch}.img"
  else:
    return f"http://cloud-images.ubuntu.com/{release}/current/{release}-server-cloudimg-{arch}.img"

def opensuse_cloud_url(arch):
  # NOTE: openSUSE Tumebleweed seems to not provide ARM64 cloud images,
  # so we do the cross-distro test on AMD64 only for now
  arch = { Arch.AMD64: "x86_64" }[arch]
  if LOCAL_IMG_SERVER:
    return f"http://localhost/cloudimg/openSUSE-Tumbleweed-Minimal-VM.{arch}-Cloud.qcow2"
  else:
    return f"https://download.opensuse.org/tumbleweed/appliances/openSUSE-Tumbleweed-Minimal-VM.{arch}-Cloud.qcow2"

def cvm_cloud_url(release, arch):
  arch = { Arch.AMD64: "amd64" }[arch]
  if LOCAL_IMG_SERVER:
    return f"http://localhost/cloudimg/{release}-server-cloudimg-{arch}-azure.fde.vhd.tar.gz"
  else:
    return f"http://cloud-images.ubuntu.com/azure/autopkgtest/tpm-fde/{release}/latest/{release}-server-cloudimg-{arch}-azure.fde.vhd.tar.gz"

def download_file(url, dest_path):
  resp = requests.get(url, stream=True)
  assert resp.status_code == 200, "Failed to download file"
  with open(dest_path, "wb") as f:
    for chunk in resp.iter_content(chunk_size=4096):
      f.write(chunk)

def fetch_file(url):
  resp = requests.get(url)
  assert resp.status_code == 200, "Failed to fetch file"
  return resp.content

def validate_packages(expected_package_set, package_paths):
  actual_package_set = set()

  for package_path in package_paths:
    _, filename = os.path.split(package_path)
    actual_package_set.add(filename.split("_")[0])

  if actual_package_set != expected_package_set:
    print("Invalid package set provided " \
          f"(actual {actual_package_set}, expecting {expected_package_set})",
          file=sys.stderr)
    return False

  return True

def download_packages(packages, dest_path):
  result = subprocess.run(["apt-get", "download"] + list(packages),
    cwd=dest_path,
    capture_output=not DEBUG)
  assert result.returncode == 0, "Failed to download packages"

def prepare_packages(tempdir_path, expected_package_set, cmdline_package_paths):
  if len(cmdline_package_paths) > 0:
    # Validate package set
    if not validate_packages(expected_package_set, cmdline_package_paths):
      exit(1)
    # Copy packages to tempdir
    for package_path in cmdline_package_paths:
      shutil.copy(package_path,
        os.path.join(tempdir_path, os.path.basename(package_path)))
  else:
    # Download packages
    download_packages(expected_package_set, tempdir_path)

  # List package file paths
  return [
    os.path.join(tempdir_path, package_filename)
    for package_filename in os.listdir(tempdir_path)
    if package_filename.endswith("deb")
  ]

EFI_CERT_X509_GUID = bytes((
  0xa1, 0x59, 0xc0, 0xa5, 0xe4, 0x94, 0xa7, 0x4a,
  0x87, 0xb5, 0xab, 0x15, 0x5c, 0x2b, 0xf0, 0x72))

EFI_CERT_SHA256_GUID = bytes((
  0x26, 0x16, 0xc4, 0xc1, 0x4c, 0x50, 0x92, 0x40,
  0xac, 0xa9, 0x41, 0xf9, 0x36, 0x93, 0x43, 0x28))

EFI_GLOBAL_VARIABLE_GUID = bytes((
  0x61, 0xdf, 0xe4, 0x8b, 0xca, 0x93, 0xd2, 0x11,
  0xaa, 0x0d, 0x00, 0xe0, 0x98, 0x03, 0x2b, 0x8c))

EFI_IMAGE_SECURITY_DATABASE_GUID = bytes((
  0xcb, 0xb2, 0x19, 0xd7, 0x3a, 0x3d, 0x96, 0x45,
  0xa3, 0xbc, 0xda, 0xd0, 0x0e, 0x67, 0x65, 0x6f))

SECURE_BOOT_ENABLE_GUID = bytes((
  0xc7, 0x0b, 0xa3, 0xf0, 0x08, 0xaf, 0x56, 0x45,
  0x99, 0xc4, 0x00, 0x10, 0x09, 0xc9, 0x3a, 0x44))

SHIM_LOCK_GUID = bytes((
  0x50, 0xab, 0x5d, 0x60, 0x46, 0xe0, 0x00, 0x43,
  0xab, 0xb6, 0x3d, 0xd8, 0x10, 0xdd, 0x8b, 0x23))

SIGNATURE_OWNER_GUID = bytes((
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42,
  0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42, 0x42))

def asn1_to_efi(asn1_data):
  signature_list_size = 16 + 4 + 4 + 4 + \
                          16 + len(asn1_data)
  signature_header_size = 0
  signature_size = 16 + len(asn1_data)
  return EFI_CERT_X509_GUID + \
         struct.pack("<III", signature_list_size,
                              signature_header_size,
                              signature_size) + \
         SIGNATURE_OWNER_GUID + asn1_data

def pe_to_efihash(pe_path):
  # Use pesign to get the authenticode hash of the PE
  # This pulls in pesign as yet another dependency, maybe we want to use:
  # https://git.launchpad.net/~mkukri/+git/sbsigntool/tree/src/sbhash.c
  result = subprocess.run(["pesign", "-h", "-i", pe_path], capture_output=True)
  assert result.returncode == 0
  digest = binascii.unhexlify(result.stdout.decode().split(" ")[0])
  assert len(digest) == 32

  # Create ESL
  signature_list_size = 16 + 4 + 4 + 4 + \
                          16 + len(digest)
  signature_header_size = 0
  signature_size = 16 + len(digest)
  return EFI_CERT_SHA256_GUID + \
         struct.pack("<III", signature_list_size,
                              signature_header_size,
                              signature_size) + \
         SIGNATURE_OWNER_GUID + digest

def gen_efi_signkey():
  """Generate an X509 signing key and certificate
  """
  key = crypto.PKey()
  key.generate_key(crypto.TYPE_RSA, 4096)

  pem_private_key = crypto.dump_privatekey(crypto.FILETYPE_PEM, key)

  cert = crypto.X509()
  cert.get_subject().CN = "ubuntu-boot-test"
  cert.set_serial_number(42)
  cert.gmtime_adj_notBefore(0)            # start now
  cert.gmtime_adj_notAfter(24 * 60 * 60)  # 1 day from now
  cert.set_issuer(cert.get_subject())
  cert.set_pubkey(key)
  cert.sign(key, "sha256")

  pem_certificate = crypto.dump_certificate(
    crypto.FILETYPE_PEM, cert)
  esl_certificate = asn1_to_efi(
    crypto.dump_certificate(crypto.FILETYPE_ASN1, cert))

  return (pem_private_key, pem_certificate, esl_certificate)

def is_uefica_signed(path):
  uefica_dir = os.path.join(os.path.dirname(__file__), "uefica")
  for filename in os.listdir(uefica_dir):
    uefica_path = os.path.join(uefica_dir, filename)
    result = subprocess.run(["sbverify", "--cert", uefica_path, path],
                            capture_output=not DEBUG)
    if result.returncode == 0:
      return True
  return False

def is_canonical_signed(path):
  canonical_uefi_ca_path = os.path.join(os.path.dirname(__file__), "canonical-uefi-ca.pem")
  result = subprocess.run(["sbverify", "--cert", canonical_uefi_ca_path, path], capture_output=not DEBUG)
  return result.returncode == 0

def maybe_gzip_ctx(path):
  class GzipCtx:
    def __init__(self, compressed_path):
      self._compressed_path = compressed_path
      self._uncompressed_path = f"{compressed_path}.ungz"
      with open(self._compressed_path, "rb") as compressed_file:
        with open(self._uncompressed_path, "wb") as uncompressed_file:
          uncompressed_file.write(gzip.decompress(compressed_file.read()))
      self.path = self._uncompressed_path
    def __enter__(self):
      return self
    def __exit__(self, type, value, traceback):
      with open(self._uncompressed_path, "rb") as uncompressed_file:
        with open(self._compressed_path, "wb") as compressed_file:
          compressed_file.write(gzip.compress(uncompressed_file.read()))

  class NullCtx:
    def __init__(self, path):
      self.path = path
    def __enter__(self):
      return self
    def __exit__(self, type, value, traceback):
      pass

  try:
    return GzipCtx(path)
  except:
    return NullCtx(path)

def deb_repack_ctx(tempdir, path):
  class DebRepackCtx:
    def __init__(self, path):
      assert path[-4:] == ".deb"
      self._deb_path = path
      self.dir_path = os.path.join(tempdir, os.path.basename(path)[:-4])
      result = subprocess.run(["dpkg-deb", "-R", self._deb_path, self.dir_path],
        capture_output=not DEBUG)
      assert result.returncode == 0
    def __enter__(self):
      return self
    def __exit__(self, type, value, traceback):
      # Suffix version with "+r1"
      with open(os.path.join(self.dir_path, "DEBIAN", "control"), "rb+") as cfile:
        lines = cfile.read().split(b"\n")
        for idx in range(len(lines)):
          if lines[idx].startswith(b"Version: "):
            lines[idx] = lines[idx].strip() + b"+r1"
        cfile.seek(0)
        cfile.write(b"\n".join(lines))
      # Repack deb
      result = subprocess.run(["dpkg-deb", "-b", self.dir_path, self._deb_path],
        capture_output=not DEBUG)
      assert result.returncode == 0
  return DebRepackCtx(path)

class SbatEntry:
  def __init__(self, line):
    cols = map(lambda col: col.strip(), line.split(b","))
    self.cnam = next(cols).decode()
    self.cgen = int(next(cols))
    self.vnam = next(cols).decode()
    self.pnam = next(cols).decode()
    self.vver = next(cols).decode()
    self.vurl = next(cols).decode()

class SbatSection:
  SBAT_SECTION_HDR = b"sbat,1,SBAT Version,sbat,1,https://github.com/rhboot/shim/blob/main/SBAT.md"

  def __init__(self, data: bytes):
    lines = data.split(b"\n")
    assert lines[0] == self.SBAT_SECTION_HDR
    self.entries = {}
    for line in lines[1:]:
      try:
        ent = SbatEntry(line)
        self.entries[ent.cnam] = ent
      except:
        continue

  def get_level_for(self, cnam: str) -> int:
    return self.entries[cnam].cgen

def decode_image_sbat(tempdir_path: str, path: str) -> SbatSection:
  # Extract .sbat section from binary
  result = subprocess.run(["objcopy",
                           path,
                           os.path.join(tempdir_path, "unused.efi"),
                           "--dump-section",
                           ".sbat=/dev/stdout"], capture_output=True)
  assert result.returncode == 0

  # Parse .sbat section
  return SbatSection(result.stdout)

class SbatLevel:
  def __init__(self, data: bytes):
    lines = data.split(b"\n")
    assert lines[0].startswith(b"sbat,1,")
    self.date = lines[0][7:].decode()
    self.entries = {}
    for line in lines[1:]:
      try:
        cnam, cgen = map(lambda col: col.strip(), line.split(b","))
        self.entries[cnam.decode()] = int(cgen)
      except:
        continue

  def encode(self) -> bytes:
    result = bytearray()
    result.extend(f"sbat,1,{self.date}\n".encode())
    for cnam, cgen in self.entries.items():
      result.extend(f"{cnam},{cgen}\n".encode())
    return bytes(result)

  def get_level_for(self, cnam: str) -> int:
    return self.entries[cnam]

  def set_level_for(self, cnam: str, cgen: int):
    self.entries[cnam] = cgen

def runcmd(args, assert_ok=True, cwd=None):
  if cwd:
    result = subprocess.run(args, capture_output=not DEBUG, cwd=cwd)
  else:
    result = subprocess.run(args, capture_output=not DEBUG)
  if assert_ok:
    assert result.returncode == 0, f"Failed to run {args}"

class ProcessWrapper:
  def __init__(self, args):
    kwargs = {}
    if not DEBUG:
      kwargs["stderr"] = subprocess.PIPE
      kwargs["stdout"] = subprocess.PIPE
    self._process = subprocess.Popen(args, **kwargs)

    # Register hook to kill process on exit
    atexit.register(self.kill_process)

  def kill_process(self):
    os.kill(self._process.pid, signal.SIGINT)
    self._process.wait()

    # We no longer need this hook now that the process is dead
    atexit.unregister(self.kill_process)

def github_get_latest_release(project):
  """
  This is a somewhat janky way to get the latest release of GitHub repo.
  The upside is, it requires no API token unlike the REST API way...
  """
  url = f"https://github.com/{project}/releases/latest"
  resp = requests.get(url)
  assert resp.status_code == 200, f"Failed to GET {url}"
  if not resp.history:
    assert False, f"GET of {url} was not redirected"
  return resp.url.split("/")[-1]

def github_download_release_asset(project, release, binary, dest_path):
  if release == "latest":
    release = github_get_latest_release(project)
  download_file(f"https://github.com/{project}/releases/download/{release}/{binary}", dest_path)



#
# ubuntu-boot-test: vm.py: Virtual machine manager
#
# Copyright (C) 2023 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pexpect import fdpexpect
from pyuefivars import UEFIVar
from pyuefivars.edk2 import EDK2UEFIVarStore
from ubuntu_boot_test.config import *
from ubuntu_boot_test.util import *
import atexit
import binascii
import os
import re
import select
import shutil
import signal
import socket
import subprocess
import tempfile

# VM configuration based on architecture and firmware type
vmconfig = {
  (Arch.AMD64, Firmware.UEFI): {
    "cmd": "qemu-system-x86_64",
    "machine": "q35",
    "cpu": "qemu64",
    "ram": "512",
    "fw_code": "/usr/share/OVMF/OVMF_CODE_4M.ms.fd",
    "fw_vars_template": "/usr/share/OVMF/OVMF_VARS_4M.ms.fd"
  },
  (Arch.AMD64, Firmware.PCBIOS): {
    "cmd": "qemu-system-x86_64",
    "machine": "q35",
    "cpu": "qemu64",
    "ram": "512"
  },
  (Arch.ARM64, Firmware.UEFI): {
    "cmd": "qemu-system-aarch64",
    "machine": "virt",
    "cpu": "cortex-a72",
    "ram": "512",
    "fw_code": "/usr/share/AAVMF/AAVMF_CODE.ms.fd",
    "fw_vars_template": "/usr/share/AAVMF/AAVMF_VARS.ms.fd"
  },
  (Arch.PPC64EL, Firmware.IEEE1275): {
    "cmd": "qemu-system-ppc64el",
    "machine": "pseries",
    "cpu": "power10",
    "ram": "512",
  }
}

# If the VM doesn't reach a desired state from boot in half an hour, then error out
vm_waittimeout = 1800

def ensure_vsock_accessible():
  lsmod_result = subprocess.run(["lsmod"], stdout=subprocess.PIPE)
  if b"vhost_vsock" not in lsmod_result.stdout:
    if os.getuid() != 0:
      assert False, "Need root to load vhost_vsock"
    if subprocess.run(["modprobe", "vhost_vsock"]).returncode != 0:
      assert False, "Failed to load vhost_vsock"

  if not os.access("/dev/vhost-vsock", os.R_OK | os.W_OK):
    assert False, "/dev/vhost-vsock is not accessible"

def create_cloud_init_seed(dest_path):
  cidata_path = os.path.join(os.path.dirname(__file__), "cidata")
  result = subprocess.run(["genisoimage", "-joliet", "-rock", "-quiet",
    "-output", dest_path, "-volid", "cidata", cidata_path],
    capture_output=not DEBUG)
  assert result.returncode == 0

class VirtualMachine:
  def __init__(self, tempdir, image_url, arch, firmware):
    if (arch, firmware) not in vmconfig:
      assert False, f"Unsupported configuration {arch} {firmware}"

    ensure_vsock_accessible()

    self._vmdir = tempdir

    self._arch = arch
    self._firmware = firmware
    self._config = vmconfig[(arch, firmware)]

    if self._firmware == Firmware.UEFI:
      self._uefi_vars_path = os.path.join(self._vmdir, "UEFI_VARS.fd")
      shutil.copy(self._config["fw_vars_template"], self._uefi_vars_path)
      # Disable 5 second boot timeout
      self.write_efivar(EFI_GLOBAL_VARIABLE_GUID, "Timeout", b"\x00\x00")

    self._disk_image_path = os.path.join(self._vmdir, "disk.img")
    if image_url is not None:
      download_file(image_url, self._disk_image_path)

    self._cloud_init_seed_path = os.path.join(self._vmdir, "cloud-init.seed")
    create_cloud_init_seed(self._cloud_init_seed_path)

    self._host_vsock = socket.socket(socket.AF_VSOCK, socket.SOCK_STREAM)
    self._host_vsock.bind((socket.VMADDR_CID_HOST, 4444))
    self._host_vsock.listen(10)
    self._host_vsock.settimeout(vm_waittimeout)

    self._monitor_socket = os.path.join(self._vmdir, "monitor")
    self._serial_socket = os.path.join(self._vmdir, "serial")

    if DEBUG:
      print(f"Created VM directory {self._vmdir}")
      print(f"Using QEMU {self._config['cmd']}")
      print(f"Using QEMU machine type {self._config['machine']}")
      if self._firmware == Firmware.UEFI:
        print(f"Using code firmware volume {self._config['fw_code']}")
        print(f"Created variable firmware volume {self._uefi_vars_path}")
      print(f"Created disk image {self._disk_image_path}")
      print(f"Created cloud-init seed {self._cloud_init_seed_path}")

  def replace_image(self, image_url):
    download_file(image_url, self._disk_image_path)

  def dump_efivars(self):
    # Read current variable store
    with open(self._uefi_vars_path, "rb") as f:
      vars_data = f.read()
    # Find existing variable
    varstore = EDK2UEFIVarStore(vars_data)
    for var in varstore.vars:
      print(f"{binascii.hexlify(var.guid).decode()}:{var.name} = {binascii.hexlify(var.data).decode()}")

  def read_efivar(self, guid, name):
    # Read current variable store
    with open(self._uefi_vars_path, "rb") as f:
      vars_data = f.read()
      vars_length = f.tell()
    # Find variable
    varstore = EDK2UEFIVarStore(vars_data)
    for var in varstore.vars:
      if var.guid == guid and var.name == name:
        return var.data
    return None

  def write_efivar(self, guid, name, data, append=True):
    # Read current variable store
    with open(self._uefi_vars_path, "rb") as f:
      vars_data = f.read()
      vars_length = f.tell()

    # Find existing variable
    varstore = EDK2UEFIVarStore(vars_data)
    found_var = None
    for var in varstore.vars:
      if var.guid == guid and var.name == name:
        found_var = var
        break

    if found_var is None:
      # Create new variable
      varstore.vars.append(UEFIVar(name, data, guid, 3))
    else:
      if data is None:
        # Delete variable
        varstore.vars.remove(found_var)
      elif append:
        # Append data
        var.data += data
      else:
        # Assign data
        var.data = data

    # Dump variable store back out
    with open(self._uefi_vars_path, "wb") as f:
      # FIXME: this is a workaround the incorrect length
      # value used by the bytes() function of EDK2UEFIVarStore
      # This means that the original file length gets lost and
      # we must manually pad the thing
      varstore.filelen = varstore.length
      f.write(bytes(varstore))
      if f.tell() < vars_length:
        f.write(b"\xff" * (vars_length - f.tell()))

  def start(self, ephemeral_snapshot=False, wait=True, tapname=None, tpm=None):
    if kvm_supported(self._arch):
      accel = "kvm"
      cpu = "host"
    else:
      accel = "tcg"
      cpu = self._config["cpu"]

    opts = [self._config["cmd"],
      "-m", self._config["ram"],
      "-M", f"{self._config['machine']},accel={accel}",
      "-cpu", cpu,
      "-nographic",
      "-vga", "none",
      "-monitor", f"unix:{self._monitor_socket},server,nowait",
      "-serial", f"unix:{self._serial_socket},server,nowait",
      "-device", "virtio-rng-pci",
      "-device", "vhost-vsock-pci,guest-cid=3",
      # First iface for network access and SSH forwarding to host
      "-netdev", "user,id=net0,hostfwd=tcp::2222-:22",
      "-device", "virtio-net-pci,romfile=,netdev=net0",
      # Second iface for netbooting
      "-netdev", "tap,id=net1,script=no,downscript=no" + (f",ifname={tapname}" if tapname else ""),
      "-device", "virtio-net-pci,mac=00:00:00:00:00:01,netdev=net1" + \
        (",bootindex=0" if tapname and self._firmware == Firmware.PCBIOS else ",romfile="),
      # NOTE: this is useful for testing DEBUG builds of OVMF, so leaving
      #  it here, but commented out
      # "-debugcon", "file:/dev/stdout",
      # "-global", "isa-debugcon.iobase=0x402"
    ]

    if tpm is not None:
      opts += [
        "-chardev", f"socket,id=chrtpm,path={tpm.ctrlsock()}",
        "-tpmdev", "emulator,id=tpm0,chardev=chrtpm",
        "-device", "tpm-tis,tpmdev=tpm0",
      ]

    if self._firmware == Firmware.UEFI:
      opts += [
        "-drive", f"file={self._config['fw_code']},if=pflash,format=raw,unit=0,readonly=on",
        "-drive", f"file={self._uefi_vars_path},if=pflash,format=raw,unit=1"
      ]

    opts += [
      "-drive", f"file={self._disk_image_path},if=none,id=hdd",
      "-device", f"virtio-blk-pci,drive=hdd,serial=0",
      "-drive", f"file={self._cloud_init_seed_path},if=virtio,format=raw,readonly=on",
    ]

    if ephemeral_snapshot:
      opts.append("-snapshot")

    # Create VM process
    self._vm_process = subprocess.Popen(opts)

    # Make sure VM exits if we error out
    def kill_vm_process():
      try:
        os.kill(self._vm_process.pid, signal.SIGINT)
      except:
        pass

    atexit.register(kill_vm_process)

    # Wait for boot if requested
    if wait:
      self.waitboot()

  def reboot(self, wait=True):
    self.run_cmd(["reboot"], assert_ok=False)
    if wait:
      self.waitboot()

  def monitor_cmd(self, cmd):
    while not os.access(self._monitor_socket, os.F_OK):
      pass
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as fd:
      fd.connect(self._monitor_socket)
      fd.send(cmd)

  def reset(self, wait=True):
    self.monitor_cmd(b"system_reset\n")
    if wait:
      self.waitboot()

  def shutdown(self, wait=True):
    self.run_cmd(["poweroff"], assert_ok=False)
    if wait:
      self.waitshutdown()

  def forceshutdown(self, wait=True):
    os.kill(self._vm_process.pid, signal.SIGINT)
    if wait:
      self.waitshutdown()

  def waitboot(self):
    # Wait for serial to come up
    while not os.access(self._serial_socket, os.F_OK):
      pass
    # Wait for serial data *or* notifier connection
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as serfd:
      serfd.connect(self._serial_socket)
      while True:
        rlist, _, _ = select.select((serfd, self._host_vsock,), (), ())
        if DEBUG and serfd in rlist:
          # Dump out boot console filtering escapes
          s = serfd.recv(4096)
          sys.stdout.buffer.write(b"".join(bytes((c,)) for c in s if c == ord('\n') or chr(c).isprintable()))
        if self._host_vsock in rlist:
          connfd, _ = self._host_vsock.accept()
          assert connfd.recv(3) == b"RDY"
          connfd.close()
          # No need to wait any longer
          return

  def waitserial(self, data):
    # Wait for serial to come up
    while not os.access(self._serial_socket, os.F_OK):
      pass
    # Wait for data
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as serfd:
      serfd.connect(self._serial_socket)
      serp = fdpexpect.fdspawn(serfd, timeout=vm_waittimeout)
      serp.expect(re.escape(data))

  def expectserial(self, handler):
    # Wait for serial to come up
    while not os.access(self._serial_socket, os.F_OK):
      pass
    # Launch expect session
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as serfd:
      serfd.connect(self._serial_socket)
      serp = fdpexpect.fdspawn(serfd, timeout=vm_waittimeout)
      handler(serp)

  def waitshutdown(self):
    self._vm_process.wait()

  def run_cmd(self, args, assert_ok=True):
    result = subprocess.run(["ssh",
      "-o", "StrictHostKeyChecking=no",
      "-o", "UserKnownHostsFile=/dev/null",
      "-p", "2222",
      "root@localhost" ] + args, capture_output=True)
    if DEBUG:
      print(result.stdout.decode())
      print(result.stderr.decode(), file=sys.stderr)
    if assert_ok:
      assert result.returncode == 0
    return result.stdout.decode()

  def copy_files(self, src_file_paths, dest_dir_path):
    result = subprocess.run(["scp",
      "-o", "StrictHostKeyChecking=no",
      "-o", "UserKnownHostsFile=/dev/null",
      "-P", "2222"] +
      src_file_paths +
      [f"root@localhost:{dest_dir_path}"], capture_output=not DEBUG)
    assert result.returncode == 0

  def get_files(self, src_file_paths, dest_dir_path):
    result = subprocess.run(["scp",
      "-o", "StrictHostKeyChecking=no",
      "-o", "UserKnownHostsFile=/dev/null",
      "-P", "2222"] +
      [ f"root@localhost:{src_file_path}" for src_file_path in src_file_paths ] +
      [ dest_dir_path ], capture_output=not DEBUG)
    assert result.returncode == 0

  def remote_file(self, remote_path):
    class RemoteFileContext:
      """Context manager for modifying remote files
      """
      def __init__(self, vm, remote_path, tempdir_path):
        self._vm = vm
        self._remote_path = remote_path
        self.local_path = os.path.join(
          tempdir_path, os.path.split(remote_path)[-1])
        self._vm.get_files([self._remote_path], self.local_path)

      def __enter__(self):
        return self

      def __exit__(self, type, value, traceback):
        self._vm.copy_files([self.local_path], self._remote_path)

    return RemoteFileContext(self, remote_path, self._vmdir)

  def disablesb(self):
    self.write_efivar(SECURE_BOOT_ENABLE_GUID, "SecureBootEnable", b"\x00", append=False)

  def enablesb(self):
    self.write_efivar(SECURE_BOOT_ENABLE_GUID, "SecureBootEnable", b"\x01", append=False)



#
# ubuntu-boot-test: tpm.py: swtpm wrapper
#
# Copyright (C) 2025 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ubuntu_boot_test.config import *
from ubuntu_boot_test.util import *
import os

class TPM:
  def __init__(self, tempdir):
    self._tempdir = tempdir
    self._pidfile = os.path.join(tempdir, "swtpm.pid")
    self._statedir = os.path.join(tempdir, "swtpm.state")
    os.makedirs(self._statedir, exist_ok=True)
    self._sock = os.path.join(tempdir, "swtpm.socket")
    self._ctrlsock = os.path.join(tempdir, "swtpm.socket.ctrl")
    self.run_process()

  def run_process(self):
    if hasattr(self, "_process"):
      # Kill previous process if present
      self._process.kill_process()
      # Wait for it to go away
      while os.access(self._pidfile, os.O_RDONLY):
        pass

    # Create process
    self._process = ProcessWrapper(["swtpm",
      "socket", "--tpm2",
      "--pid", f"file={self._pidfile}",
      "--tpmstate", f"dir={self._statedir}",
      "--server", f"type=unixio,path={self._sock},mode=0666",
      "--ctrl", f"type=unixio,path={self._ctrlsock},mode=0666",
      "--flags", "not-need-init,startup-clear",
    ])
    # Wait for it to start
    while not os.access(self._pidfile, os.O_RDONLY):
      pass
    # Send startup command
    # NOTE: This shouldnt be needed and should be done by the BIOS
    # but something is in a bad state without it.
    self.startup()

  def sock(self):
    return self._sock

  def ctrlsock(self):
    return self._ctrlsock

  def generate_srk(self, outdir):
    runcmd(["tpm2_createprimary",
      "-T", f"swtpm:path={self._sock}",
      "-c", os.path.join(outdir, "srk.ctx"),
    ])
    runcmd(["tpm2_readpublic",
      "-T", f"swtpm:path={self._sock}",
      "-c", os.path.join(outdir, "srk.ctx"),
      "-o", os.path.join(outdir, "srk.pub"),
    ])

  def startup(self):
    runcmd(["tpm2_startup", "-T", f"swtpm:path={self._sock}", "-c"])



# ubuntu-boot-test: cmd_stubble.py: stubble test
#
# Copyright (C) 2025 Canonical, Ltd.
# Author: Mate Kukri <mate.kukri@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ubuntu_boot_test.config import *
from ubuntu_boot_test.net import VirtualNetwork
from ubuntu_boot_test.util import *
from ubuntu_boot_test.vm import VirtualMachine
import json
import os
import shutil
import subprocess
import tempfile

def register(subparsers):
  parser = subparsers.add_parser("stubble", description="stubble test")

  parser.add_argument("-r", "--release", required=True,
    help="Guest Ubuntu release")
  parser.add_argument("-a", "--arch", required=True, type=Arch,
    help="Guest architecture")
  parser.add_argument("packages", nargs="*",
    help="List of packages to install (instead of apt-get download)")

def execute(args):
  TEMPDIR = tempfile.TemporaryDirectory("")

  PACKAGE_SETS = {
    Arch.AMD64: set(("stubble", "systemd-ukify", "python3-pefile", "python3-zstandard", )),
    Arch.ARM64: set(("stubble", "systemd-ukify", "python3-pefile", "python3-zstandard", )),
  }

  # Package paths to install
  package_paths = prepare_packages(TEMPDIR.name, PACKAGE_SETS[args.arch], args.packages)
  # Create virtual machine
  vm = VirtualMachine(TEMPDIR.name, ubuntu_cloud_url(args.release, args.arch), args.arch, Firmware.UEFI)

  def gen_sb_key(guid, name):
    if os.path.isfile(f"{name}.key"):
        return
    pem_priv, pem_cert, esl_cert = gen_efi_signkey()
    with open(os.path.join(TEMPDIR.name, f"{name}.key"), "wb") as f:
      f.write(pem_priv)
    with open(os.path.join(TEMPDIR.name, f"{name}.pem"), "wb") as f:
      f.write(pem_cert)
    vm.write_efivar(guid, name, esl_cert, append=True)

  gen_sb_key(SHIM_LOCK_GUID, "MokList")

  def sbsign_file(with_key, path):
    result = subprocess.run(["sbsign",
      "--key", os.path.join(TEMPDIR.name, f"{with_key}.key"),
      "--cert", os.path.join(TEMPDIR.name, f"{with_key}.pem"),
      "--output", path, path], capture_output=not DEBUG)
    assert result.returncode == 0, f"Failed to sign {path}"

  def installnew():
    # Copy packages to VM
    vm.copy_files(package_paths, "/tmp/")
    # Install packages
    vm.run_cmd(["apt", "install", "--yes", "/tmp/*.deb"])

  def wrapkernel():
    # Wrap kernel in stubble
    vm.run_cmd(["bash", "-c", "\"zcat -f /boot/vmlinuz > /boot/vmlinuz.tmp\""])
    vm.run_cmd([
      "ukify", "build", "--linux=/boot/vmlinuz.tmp", "--stub=/usr/lib/stubble/stubble.efi", "--output=/boot/vmlinuz.efi"
    ])
    # Replace original kernel
    real_kernel_path = vm.run_cmd(["readlink", "-f", "/boot/vmlinuz"]).strip()
    vm.run_cmd(["mv", "/boot/vmlinuz.efi", real_kernel_path])

  def checkunsigned():
    # Reboot and wait for error from GRUB
    # To make sure it does not boot before signing
    vm.reboot(wait=False)
    vm.waitserial(b"error: bad shim lock signature.") # FIXME: this will change for loader protocol

    # Force shutdown VM
    vm.forceshutdown()

  def signstubble():
    real_kernel_path = vm.run_cmd(["readlink", "-f", "/boot/vmlinuz"]).strip()
    with vm.remote_file(real_kernel_path) as rf:
      sbsign_file("MokList", rf.local_path)

  TASKS = [
    (vm.start,      "Boot and provision image"),
    (installnew,    "Install new packages"),
    (wrapkernel,    "Wrap kernel in stubble"),
    (checkunsigned, "Ensure unsinged stubble does not boot"),
    (vm.disablesb,  "Disable SecureBoot"),
    (vm.start,      "Boot virtual machine"),
    (signstubble,   "Sign stubble image"),
    (vm.shutdown,   "Shut down virtual machine"),
    (vm.enablesb,   "Enable SecureBoot"),
    (vm.start,      "Boot with signed stubble image"),
    (vm.shutdown,   "Shut down virtual machine"),
  ]

  for do_task, msg in TASKS:
    do_task()
    print(f"{msg} OK")
